package be.iminds.ilabt.jfed.highlevel.jobs;

import be.iminds.ilabt.jfed.experiment.Experiment;
import be.iminds.ilabt.jfed.experiment.ExperimentPart;
import be.iminds.ilabt.jfed.experiment.SfaExperimentPart;
import be.iminds.ilabt.jfed.experiment.tasks.ExperimentTaskStatus;
import be.iminds.ilabt.jfed.fedmon.webapi.service.json.Server;
import be.iminds.ilabt.jfed.highlevel.controller.TaskThread;
import be.iminds.ilabt.jfed.highlevel.jobs.parts.ExperimentPartController;
import be.iminds.ilabt.jfed.highlevel.jobs.parts.ExperimentPartControllerManager;
import be.iminds.ilabt.jfed.highlevel.jobs.parts.NonSfaExperimentPartController;
import be.iminds.ilabt.jfed.highlevel.jobs.sfa_parts.RegisterSliversToSlaExperimentPartStateSlice;
import be.iminds.ilabt.jfed.highlevel.jobs.sfa_parts.UpdateSfaExperimentPartUntilReadyStateSlice;
import be.iminds.ilabt.jfed.highlevel.jobs.states.JobStateFactory;
import be.iminds.ilabt.jfed.highlevel.jobs.states.ProvisionSliversOnAuthorityState;
import be.iminds.ilabt.jfed.highlevel.jobs.states.ShareLanAtAuthorityPartState;
import be.iminds.ilabt.jfed.highlevel.model.InternalState;
import be.iminds.ilabt.jfed.highlevel.tasks.HighLevelTaskFactory;
import be.iminds.ilabt.jfed.highlevel.util.AggregateManagerWrapperFactory;
import be.iminds.ilabt.jfed.lowlevel.JFedException;
import be.iminds.ilabt.jfed.lowlevel.api.user_spec.UserSpec;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.AggregateManagerWrapper;
import be.iminds.ilabt.jfed.rspec.model.ModelRspecType;
import be.iminds.ilabt.jfed.rspec.model.RspecLink;
import be.iminds.ilabt.jfed.rspec.model.RspecNode;
import be.iminds.ilabt.jfed.rspec.model.javafx_impl.FXModelRspec;
import be.iminds.ilabt.jfed.rspec.rspec_source.RequestRspecSource;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import java.util.*;
import java.util.stream.Collectors;

/**
 * User: twalcari
 * Date: 12/2/2015
 * Time: 16:43
 */
public class WaitForReadyExperimentJob extends AbstractJob<Boolean> {
    private static final Logger LOG = LoggerFactory.getLogger(WaitForReadyExperimentJob.class);

    private final JobStateFactory jobStateFactory;

    private final AggregateManagerWrapperFactory aggregateManagerWrapperFactory;
    private final ExperimentPartControllerManager epcManager;

    private Boolean success = null;

    public WaitForReadyExperimentJob(Experiment experiment,
                                     HighLevelTaskFactory hltf,
                                     TaskThread tt,
                                     JobStateFactory jobStateFactory,
                                     AggregateManagerWrapperFactory aggregateManagerWrapperFactory,
                                     ExperimentPartControllerManager epcManager) {
        super("Wait for Ready for Experiment " + experiment.getName(), experiment, hltf, tt);
        this.jobStateFactory = jobStateFactory;
        this.aggregateManagerWrapperFactory = aggregateManagerWrapperFactory;
        this.epcManager = epcManager;

        assert experiment.getRequestedEndTime() != null;
    }

    @Override
    public Boolean execute() throws Exception {
        if (!hasSfaParts()) {
            if (experiment.getExperimentSpecification() != null) {
                experiment.getExperimentSpecificationLogger().firePostRSpec(experiment.getExperimentSpecification().getRspecs().get(0), false);
                experiment.getExperimentSpecificationLogger().firePostRspecAll(false, Collections.emptyList());
            }
            return true;
        }

        boolean permanentFailure = false;

        LOG.debug("WaitForReadyExperimentJob started");

        final UpdateSfaPartsUntilReadyState updateUntilReadyState =
                new UpdateSfaPartsUntilReadyState(jobStateFactory, this);
        setAndRunState(updateUntilReadyState);

        if (updateUntilReadyState.isPermanentFailure()) {
            permanentFailure = true;
            LOG.info("WaitForPartsReadyState reported permanent failure");
        }

        if (permanentFailure) {
            if (experiment.getExperimentSpecification() != null) {
                experiment.getExperimentSpecificationLogger().firePostRSpec(
                        experiment.getExperimentSpecification().getRspecs().get(0), false);
                experiment.getExperimentSpecificationLogger().firePostRspecAll(
                        false, getExperimentClientIds());
            }

            LOG.debug("WaitForReadyExperimentJob stopped in failure");
            success = false;
            return false;
        }

        LOG.debug("WaitForReadyExperimentJob stopped successfully");

        success = true;
        return true;
    }

    @Nonnull
    private List<String> getExperimentClientIds() {
        if (experiment.getSliceOrNull() == null || experiment.getSliceOrNull().getRequestRspec() == null) {
            return Collections.emptyList();
        }
        RequestRspecSource requestRspecSource = experiment.getSliceOrNull().getRequestRspec();
        FXModelRspec model = requestRspecSource == null ?
                null :
                requestRspecSource.getModelRspec(ModelRspecType.FX);
        assert model != null;

        if (model == null) {
            return Collections.emptyList();
        }

        return model.getNodes().stream().map(RspecNode::getClientId).filter(Objects::nonNull).collect(Collectors.toList());
    }

    public Boolean isSuccess() {
        return success;
    }

    public static class UpdateSfaPartsUntilReadyState extends SlicedState<UpdateSfaExperimentPartUntilReadyStateSlice> {
        private final List<UpdateSfaExperimentPartUntilReadyStateSlice> slices;
        private final Experiment experiment;
        private final JobStateFactory jobStateFactory;
        private final Job job;

        UpdateSfaPartsUntilReadyState(@Nonnull JobStateFactory jobStateFactory, @Nonnull Job job) {
            super("Waiting for resources to become ready");

            this.jobStateFactory = jobStateFactory;
            this.job = job;
            this.experiment = job.getExperiment();

            slices = experiment.getPartsStream()
                    .filter(part -> part.getState() != InternalState.FAILED)
                    .filter(part -> part instanceof SfaExperimentPart)
                    .map(part -> new UpdateSfaExperimentPartUntilReadyStateSlice(
                            job,
                            (SfaExperimentPart) part,
                            experiment.getCheckReadyInterval(),
                            experiment.getMaxWaitUntilReady(),
                            jobStateFactory))
                    .collect(Collectors.toList());
        }


        public boolean isPermanentFailure() {
            return getSlices().stream().anyMatch(updateSlice -> updateSlice.getResult() != ExperimentTaskStatus.SUCCESS);
        }

        @Override
        public Collection<UpdateSfaExperimentPartUntilReadyStateSlice> getSlices() {
            return slices;
        }
    }






    private Collection<SfaExperimentPart> getSfaParts() {
        return experiment.getPartsStream()
                .filter(part -> part instanceof SfaExperimentPart)
                .map(SfaExperimentPart.class::cast)
                .collect(Collectors.toList());

    }

    private Collection<NonSfaExperimentPartController> getNonSfaExperimentPartControllers() {

        Set<NonSfaExperimentPartController> controllers = new HashSet<>();

        for (ExperimentPart experimentPart : experiment.getPartsListCopy()) {
            if (experimentPart instanceof SfaExperimentPart)
                continue;


            ExperimentPartController epc = epcManager.getController(experimentPart);
            if (!(epc instanceof NonSfaExperimentPartController))
                throw new RuntimeException("Expected a non-sfa ExperimentPartController");

            controllers.add((NonSfaExperimentPartController) epc);
        }

        return controllers;
    }

    private boolean hasNonSfaParts() {
        return experiment.getPartsStream().anyMatch(part -> !(part instanceof SfaExperimentPart));
    }

    private boolean hasSfaParts() {
        return experiment.getPartsStream().anyMatch(part -> (part instanceof SfaExperimentPart));
    }
}
